"""Input validation utilities for FN-DBSCAN algorithm.

This module provides functions to validate input parameters and data
for the FN-DBSCAN clustering algorithm.
"""

import numpy as np
from typing import Union, Tuple


def validate_eps(eps: float) -> None:
    """Validate epsilon parameter.

    Parameters
    ----------
    eps : float
        Epsilon (maximum neighborhood radius).

    Raises
    ------
    ValueError
        If eps is not positive.
    TypeError
        If eps is not a number.
    """
    if not isinstance(eps, (int, float, np.number)):
        raise TypeError(f"eps must be a number, got {type(eps)}")

    if not np.isfinite(eps):
        raise ValueError(f"eps must be finite, got {eps}")

    if eps <= 0:
        raise ValueError(f"eps must be > 0, got {eps}")


def validate_epsilon2(epsilon2: float) -> None:
    """Validate epsilon2 (minimum fuzzy cardinality) parameter.

    Parameters
    ----------
    epsilon2 : float
        Minimum fuzzy cardinality for core points (ε₂ in the paper).

    Raises
    ------
    ValueError
        If epsilon2 is not positive.
    TypeError
        If epsilon2 is not a number.
    """
    if not isinstance(epsilon2, (int, float, np.number)):
        raise TypeError(
            f"epsilon2 must be a number, got {type(epsilon2)}"
        )

    if not np.isfinite(epsilon2):
        raise ValueError(f"epsilon2 must be finite, got {epsilon2}")

    if epsilon2 <= 0:
        raise ValueError(f"epsilon2 must be > 0, got {epsilon2}")


# Backward compatibility alias
def validate_min_cardinality(min_cardinality: float) -> None:
    """Deprecated. Use validate_epsilon2 instead."""
    import warnings
    warnings.warn(
        "validate_min_cardinality is deprecated. Use validate_epsilon2 instead.",
        DeprecationWarning,
        stacklevel=2
    )
    validate_epsilon2(min_cardinality)


def validate_fuzzy_function(fuzzy_function: str) -> None:
    """Validate fuzzy function parameter.

    Parameters
    ----------
    fuzzy_function : str
        Type of fuzzy membership function.

    Raises
    ------
    ValueError
        If fuzzy_function is not recognized.
    TypeError
        If fuzzy_function is not a string.
    """
    if not isinstance(fuzzy_function, str):
        raise TypeError(
            f"fuzzy_function must be a string, got {type(fuzzy_function)}"
        )

    valid_functions = {'linear', 'exponential', 'trapezoidal'}
    if fuzzy_function not in valid_functions:
        raise ValueError(
            f"fuzzy_function must be one of {valid_functions}, "
            f"got '{fuzzy_function}'"
        )


def validate_metric(metric: str) -> None:
    """Validate distance metric parameter.

    Parameters
    ----------
    metric : str
        Distance metric name.

    Raises
    ------
    TypeError
        If metric is not a string or callable.

    Notes
    -----
    This function performs basic validation. The actual metric validation
    is delegated to sklearn's NearestNeighbors class.
    """
    if not isinstance(metric, (str, callable)):
        raise TypeError(
            f"metric must be a string or callable, got {type(metric)}"
        )


def validate_data(X: np.ndarray) -> Tuple[np.ndarray, int, int]:
    """Validate input data array.

    Parameters
    ----------
    X : array-like
        Input data array.

    Returns
    -------
    X_validated : ndarray
        Validated and converted data array.
    n_samples : int
        Number of samples.
    n_features : int
        Number of features.

    Raises
    ------
    ValueError
        If X contains NaN or Inf values.
        If X is not 2D.
        If X has no samples.
    TypeError
        If X cannot be converted to ndarray.
    """
    # Convert to numpy array if needed
    try:
        X = np.asarray(X, dtype=float)
    except (ValueError, TypeError) as e:
        raise TypeError(f"X must be convertible to numpy array: {e}")

    # Check dimensionality
    if X.ndim == 1:
        # Single sample or single feature - reshape
        X = X.reshape(-1, 1) if len(X) > 1 else X.reshape(1, -1)

    if X.ndim != 2:
        raise ValueError(
            f"X must be a 2D array, got {X.ndim}D array with shape {X.shape}"
        )

    n_samples, n_features = X.shape

    # Check for empty data
    if n_samples == 0:
        raise ValueError("X has 0 samples")

    if n_features == 0:
        raise ValueError("X has 0 features")

    # Check for NaN/Inf
    if not np.all(np.isfinite(X)):
        raise ValueError("X contains NaN or Inf values")

    return X, n_samples, n_features


def validate_fit_params(eps: float, epsilon2: float,
                        fuzzy_function: str, metric: str) -> None:
    """Validate all fit parameters.

    Parameters
    ----------
    eps : float
        Epsilon (maximum neighborhood radius).
    epsilon2 : float
        Minimum fuzzy cardinality for core points.
    fuzzy_function : str
        Type of fuzzy membership function.
    metric : str
        Distance metric name.

    Raises
    ------
    ValueError
        If any parameter is invalid.
    TypeError
        If any parameter has wrong type.
    """
    validate_eps(eps)
    validate_epsilon2(epsilon2)
    validate_fuzzy_function(fuzzy_function)
    validate_metric(metric)



"""Tests for FN_DBSCAN main algorithm."""

import numpy as np
import pytest
from sklearn.datasets import make_blobs, make_moons

from fn_dbscan import FN_DBSCAN


class TestFNDBSCANBasic:
    """Basic tests for FN_DBSCAN algorithm."""

    def test_initialization(self):
        """Test FN_DBSCAN initialization."""
        model = FN_DBSCAN(eps=0.5, min_fuzzy_neighbors=3, fuzzy_function='linear')
        assert model.eps == 0.5
        assert model.min_fuzzy_neighbors == 3
        assert model.fuzzy_function == 'linear'
        assert model.metric == 'euclidean'

    def test_repr(self):
        """Test string representation."""
        model = FN_DBSCAN(eps=0.5, min_fuzzy_neighbors=3)
        repr_str = repr(model)
        assert 'FN_DBSCAN' in repr_str
        assert 'eps=0.5' in repr_str
        assert 'min_fuzzy_neighbors=3' in repr_str

    def test_simple_clustering(self):
        """Test clustering on simple 2-cluster dataset."""
        # Create two well-separated blobs
        X, y_true = make_blobs(
            n_samples=50,
            centers=2,
            n_features=2,
            cluster_std=0.5,
            center_box=(0, 10),
            random_state=42
        )

        model = FN_DBSCAN(eps=2.0, min_fuzzy_neighbors=3)
        labels = model.fit_predict(X)

        # Should find 2 clusters
        assert model.n_clusters_ == 2
        assert len(labels) == 50
        assert set(labels) <= {-1, 0, 1}  # Only -1, 0, 1 should appear

    def test_fit_returns_self(self):
        """Test that fit returns self."""
        X = np.array([[1, 2], [2, 2], [2, 3]])
        model = FN_DBSCAN()
        result = model.fit(X)
        assert result is model

    def test_fit_predict_consistency(self):
        """Test that fit and fit_predict give same results."""
        X = np.array([[1, 2], [2, 2], [2, 3], [8, 7], [8, 8]])

        model1 = FN_DBSCAN(eps=3, min_fuzzy_neighbors=2)
        model1.fit(X)

        model2 = FN_DBSCAN(eps=3, min_fuzzy_neighbors=2)
        labels = model2.fit_predict(X)

        np.testing.assert_array_equal(model1.labels_, labels)

    def test_attributes_after_fit(self):
        """Test that required attributes exist after fit."""
        X = np.array([[1, 2], [2, 2], [2, 3]])
        model = FN_DBSCAN(eps=3, min_fuzzy_neighbors=2)
        model.fit(X)

        assert hasattr(model, 'labels_')
        assert hasattr(model, 'core_sample_indices_')
        assert hasattr(model, 'n_clusters_')

        assert len(model.labels_) == 3
        assert isinstance(model.core_sample_indices_, np.ndarray)
        assert isinstance(model.n_clusters_, (int, np.integer))


class TestFNDBSCANFuzzyFunctions:
    """Test FN_DBSCAN with different fuzzy functions."""

    def setup_method(self):
        """Set up test data."""
        self.X = np.array([
            [1, 2], [2, 2], [2, 3],  # Close cluster
            [8, 7], [8, 8], [25, 80]  # Another cluster + noise
        ])

    def test_linear_function(self):
        """Test with linear fuzzy function."""
        model = FN_DBSCAN(eps=3, min_fuzzy_neighbors=2, fuzzy_function='linear')
        model.fit(self.X)

        assert model.n_clusters_ >= 0
        assert len(model.labels_) == 6

    def test_exponential_function(self):
        """Test with exponential fuzzy function."""
        model = FN_DBSCAN(eps=3, min_fuzzy_neighbors=2, fuzzy_function='exponential')
        model.fit(self.X)

        assert model.n_clusters_ >= 0
        assert len(model.labels_) == 6

    def test_trapezoidal_function(self):
        """Test with trapezoidal fuzzy function."""
        model = FN_DBSCAN(eps=3, min_fuzzy_neighbors=2, fuzzy_function='trapezoidal')
        model.fit(self.X)

        assert model.n_clusters_ >= 0
        assert len(model.labels_) == 6

    def test_different_functions_may_differ(self):
        """Test that different fuzzy functions can produce different results."""
        labels_linear = FN_DBSCAN(
            eps=3, min_fuzzy_neighbors=2.5, fuzzy_function='linear'
        ).fit_predict(self.X)

        labels_exp = FN_DBSCAN(
            eps=3, min_fuzzy_neighbors=2.5, fuzzy_function='exponential'
        ).fit_predict(self.X)

        # Results may differ (but not necessarily)
        # Just check they both work
        assert len(labels_linear) == len(labels_exp) == 6


class TestFNDBSCANEdgeCases:
    """Test edge cases for FN_DBSCAN."""

    def test_empty_dataset(self):
        """Test with empty dataset."""
        X = np.array([]).reshape(0, 2)
        model = FN_DBSCAN()

        with pytest.raises(ValueError, match="0 samples"):
            model.fit(X)

    def test_single_point(self):
        """Test with single point."""
        X = np.array([[1, 2]])
        model = FN_DBSCAN(eps=0.5, min_fuzzy_neighbors=3)
        model.fit(X)

        # Single point cannot be core (cardinality = 1.0 < 3)
        assert model.labels_[0] == -1
        assert model.n_clusters_ == 0

    def test_two_identical_points(self):
        """Test with two identical points."""
        X = np.array([[1, 2], [1, 2]])
        model = FN_DBSCAN(eps=0.5, min_fuzzy_neighbors=2)
        model.fit(X)

        # Both points have cardinality 2.0, should form cluster
        assert model.n_clusters_ == 1
        assert model.labels_[0] == model.labels_[1]
        assert model.labels_[0] != -1

    def test_all_noise(self):
        """Test when all points are noise."""
        X = np.array([[1, 2], [10, 20], [100, 200]])
        model = FN_DBSCAN(eps=0.1, min_fuzzy_neighbors=5)
        model.fit(X)

        assert model.n_clusters_ == 0
        assert np.all(model.labels_ == -1)

    def test_single_cluster(self):
        """Test when all points form one cluster."""
        X = np.array([[1, 2], [1.5, 2.5], [2, 2], [2.5, 2.5]])
        model = FN_DBSCAN(eps=2.0, min_fuzzy_neighbors=2)
        model.fit(X)

        # All points should be in same cluster
        assert model.n_clusters_ == 1
        assert len(np.unique(model.labels_)) == 1
        assert model.labels_[0] == 0

    def test_min_cardinality_one(self):
        """Test with min_fuzzy_neighbors=1 (everything is core)."""
        X = np.array([[1, 2], [10, 20], [100, 200]])
        model = FN_DBSCAN(eps=1.0, min_fuzzy_neighbors=1.0)
        model.fit(X)

        # Each point has at least cardinality 1.0 (itself)
        # So all should be core, but may not be connected
        assert len(model.core_sample_indices_) == 3


class TestFNDBSCANNonConvex:
    """Test FN_DBSCAN on non-convex datasets."""

    def test_moons_dataset(self):
        """Test on sklearn's moons dataset."""
        X, y_true = make_moons(n_samples=100, noise=0.1, random_state=42)

        model = FN_DBSCAN(eps=0.3, min_fuzzy_neighbors=2.5)
        model.fit(X)

        # Should find approximately 2 clusters
        assert 1 <= model.n_clusters_ <= 3

    def test_deterministic(self):
        """Test that algorithm is deterministic."""
        X, _ = make_blobs(n_samples=50, random_state=42)

        model1 = FN_DBSCAN(eps=2.0, min_fuzzy_neighbors=3)
        labels1 = model1.fit_predict(X)

        model2 = FN_DBSCAN(eps=2.0, min_fuzzy_neighbors=3)
        labels2 = model2.fit_predict(X)

        np.testing.assert_array_equal(labels1, labels2)


class TestFNDBSCANInputValidation:
    """Test input validation for FN_DBSCAN."""

    def test_invalid_eps(self):
        """Test with invalid eps values."""
        X = np.array([[1, 2], [2, 3]])

        with pytest.raises((ValueError, TypeError)):
            FN_DBSCAN(eps=0).fit(X)

        with pytest.raises((ValueError, TypeError)):
            FN_DBSCAN(eps=-1).fit(X)

        with pytest.raises((ValueError, TypeError)):
            FN_DBSCAN(eps=np.inf).fit(X)

    def test_invalid_min_cardinality(self):
        """Test with invalid min_fuzzy_neighbors values."""
        X = np.array([[1, 2], [2, 3]])

        with pytest.raises((ValueError, TypeError)):
            FN_DBSCAN(min_fuzzy_neighbors=0).fit(X)

        with pytest.raises((ValueError, TypeError)):
            FN_DBSCAN(min_fuzzy_neighbors=-1).fit(X)

    def test_invalid_fuzzy_function(self):
        """Test with invalid fuzzy function."""
        X = np.array([[1, 2], [2, 3]])

        with pytest.raises((ValueError, TypeError)):
            FN_DBSCAN(fuzzy_function='invalid').fit(X)

        with pytest.raises((ValueError, TypeError)):
            FN_DBSCAN(fuzzy_function=123).fit(X)

    def test_nan_in_data(self):
        """Test with NaN in data."""
        X = np.array([[1, 2], [np.nan, 3]])
        model = FN_DBSCAN()

        with pytest.raises(ValueError, match="NaN"):
            model.fit(X)

    def test_inf_in_data(self):
        """Test with Inf in data."""
        X = np.array([[1, 2], [np.inf, 3]])
        model = FN_DBSCAN()

        with pytest.raises(ValueError, match="Inf"):
            model.fit(X)

    def test_1d_data(self):
        """Test with 1D data."""
        X = np.array([1, 2, 3, 4, 5])
        model = FN_DBSCAN()
        # Should reshape to 2D automatically
        model.fit(X)
        assert model.labels_.shape == (5,)


class TestFNDBSCANCore:
    """Test core point detection."""

    def test_core_sample_indices(self):
        """Test that core_sample_indices_ is correct."""
        X = np.array([
            [0, 0], [1, 0], [0, 1],  # Dense region
            [10, 10]  # Isolated point
        ])

        model = FN_DBSCAN(eps=2.0, min_fuzzy_neighbors=2)
        model.fit(X)

        # First 3 points should be core
        assert len(model.core_sample_indices_) >= 1

        # Core samples should not include noise
        for idx in model.core_sample_indices_:
            assert model.labels_[idx] != -1

    def test_noise_point_becomes_border(self):
        """Test that noise points can become border points."""
        # Create scenario where a point is initially noise but
        # gets absorbed by expanding cluster
        X = np.array([
            [0, 0], [1, 0], [2, 0],  # Core points
            [3, 0]  # Border point (not enough neighbors to be core)
        ])

        model = FN_DBSCAN(eps=1.5, min_fuzzy_neighbors=1.6)
        model.fit(X)

        # All points should be in same cluster (no noise)
        # because border point is reachable from core
        assert np.all(model.labels_ >= 0)



"""Test edge cases and corner cases for FN-DBSCAN.

This module contains tests for all edge cases documented in the
technical specification.
"""

import numpy as np
import pytest

from fn_dbscan import FN_DBSCAN


class TestInputEdgeCases:
    """Test edge cases related to input data."""

    def test_edge_case_1_empty_dataset(self):
        """Edge Case 1: Empty dataset."""
        X = np.array([]).reshape(0, 2)
        model = FN_DBSCAN(eps=0.5, min_fuzzy_neighbors=3)

        with pytest.raises(ValueError, match="0 samples"):
            model.fit(X)

    def test_edge_case_2_single_point(self):
        """Edge Case 2: Single point."""
        X = np.array([[1, 2]])
        model = FN_DBSCAN(eps=0.5, min_fuzzy_neighbors=3)
        model.fit(X)

        # Cardinality = 1.0 < 3, so noise
        assert model.labels_[0] == -1
        assert model.n_clusters_ == 0

    def test_edge_case_3_two_identical_points(self):
        """Edge Case 3: Two identical points."""
        X = np.array([[1, 2], [1, 2]])
        model = FN_DBSCAN(eps=0.5, min_fuzzy_neighbors=2)
        model.fit(X)

        # Both points have cardinality 2.0
        assert model.n_clusters_ == 1
        assert model.labels_[0] == model.labels_[1]
        assert model.labels_[0] == 0

    def test_edge_case_4_min_cardinality_one(self):
        """Edge Case 4: MinCard = 1."""
        X = np.array([[1, 2], [10, 20], [100, 200]])
        model = FN_DBSCAN(eps=1.0, min_fuzzy_neighbors=1.0)
        model.fit(X)

        # Every point is core (cardinality ≥ 1.0)
        assert len(model.core_sample_indices_) == 3

    def test_edge_case_5_eps_zero(self):
        """Edge Case 5: ε = 0 (should fail)."""
        X = np.array([[1, 2], [2, 3]])

        with pytest.raises(ValueError):
            FN_DBSCAN(eps=0).fit(X)

    def test_edge_case_6_extremely_small_eps(self):
        """Edge Case 6: Extremely small ε."""
        X = np.array([[0.0, 0.0], [0.1, 0.1], [0.2, 0.2]])
        model = FN_DBSCAN(eps=0.01, min_fuzzy_neighbors=2)
        model.fit(X)

        # No point has neighbors, all noise
        assert np.all(model.labels_ == -1)
        assert model.n_clusters_ == 0


class TestNumericalEdgeCases:
    """Test numerical edge cases."""

    def test_edge_case_7_very_small_distances(self):
        """Edge Case 7: Very small distances."""
        X = np.array([
            [0.0, 0.0],
            [1e-15, 1e-15],  # Extremely close
            [1e-10, 1e-10]
        ])
        model = FN_DBSCAN(eps=1.0, min_fuzzy_neighbors=2)
        model.fit(X)

        # Should handle small distances correctly
        assert model.n_clusters_ >= 0

    def test_edge_case_8_distance_exactly_at_eps(self):
        """Edge Case 8: Distance exactly at ε."""
        # Create points exactly eps apart
        X = np.array([[0, 0], [1, 0]])  # Distance = 1.0
        model = FN_DBSCAN(eps=1.0, min_fuzzy_neighbors=1.5, fuzzy_function='linear')
        model.fit(X)

        # For linear: membership at distance=eps is 0
        # Each point has cardinality = 1.0 (self) + 0.0 (other) = 1.0
        # 1.0 < 1.5, so both are noise
        assert np.all(model.labels_ == -1)

    def test_edge_case_9_nan_in_data(self):
        """Edge Case 9: NaN in data."""
        X = np.array([[1, 2], [np.nan, 3], [4, 5]])
        model = FN_DBSCAN()

        with pytest.raises(ValueError, match="NaN"):
            model.fit(X)

    def test_edge_case_10_inf_in_data(self):
        """Edge Case 10: Inf in data."""
        X = np.array([[1, 2], [np.inf, 3], [4, 5]])
        model = FN_DBSCAN()

        with pytest.raises(ValueError, match="Inf"):
            model.fit(X)

    def test_edge_case_cardinality_just_below_threshold(self):
        """Edge Case: Cardinality just below MinCard."""
        # Create scenario where cardinality is very close to threshold
        X = np.array([[0, 0], [1, 0], [2, 0]])

        # With linear membership and eps=1.5:
        # Point 1: self(1.0) + point0(~0.33) + point2(~0.33) ≈ 1.66
        model = FN_DBSCAN(eps=1.5, min_fuzzy_neighbors=1.67)
        model.fit(X)

        # Point 1 should not be core
        # But exact behavior depends on floating point
        assert model.n_clusters_ >= 0


class TestAlgorithmicEdgeCases:
    """Test algorithmic edge cases."""

    def test_edge_case_11_nested_clusters(self):
        """Edge Case 11: Nested clusters."""
        # Dense inner region + sparse outer region
        inner = np.random.RandomState(42).randn(20, 2) * 0.3
        outer = np.random.RandomState(43).randn(10, 2) * 3.0

        X = np.vstack([inner, outer])

        model = FN_DBSCAN(eps=1.0, min_fuzzy_neighbors=3)
        model.fit(X)

        # Should detect at least inner cluster
        assert model.n_clusters_ >= 1

    def test_edge_case_12_border_point_ambiguity(self):
        """Edge Case 12: Border point between two cores."""
        # Create two core points with a point in between
        X = np.array([
            [0, 0],   # Core 1
            [1, 0],   # Border (between two cores)
            [2, 0],   # Core 2
            [0, 0.1], # Support for core 1
            [2, 0.1]  # Support for core 2
        ])

        model = FN_DBSCAN(eps=1.5, min_fuzzy_neighbors=2)
        model.fit(X)

        # Border point should be assigned deterministically
        # to first cluster that reaches it
        middle_label = model.labels_[1]
        assert middle_label >= 0  # Not noise

    def test_edge_case_13_noise_point_reclassification(self):
        """Edge Case 13: Noise point becomes border point."""
        X = np.array([
            [0, 0], [1, 0], [2, 0],  # Core points
            [3, 0]  # Initially noise, but reachable
        ])

        model = FN_DBSCAN(eps=1.5, min_fuzzy_neighbors=1.6)
        model.fit(X)

        # Point 3 should be border point (not noise)
        # because it's reachable from core points
        assert model.labels_[3] >= 0

    def test_edge_case_14_all_noise(self):
        """Edge Case 14: All noise result."""
        # Well-separated points, impossible to cluster
        X = np.array([
            [0, 0],
            [100, 100],
            [200, 200],
            [300, 300]
        ])

        model = FN_DBSCAN(eps=1.0, min_fuzzy_neighbors=5)
        model.fit(X)

        assert model.n_clusters_ == 0
        assert np.all(model.labels_ == -1)
        assert len(model.core_sample_indices_) == 0

    def test_edge_case_15_single_cluster(self):
        """Edge Case 15: Single cluster result."""
        # Dense cluster where all points are connected
        X = np.random.RandomState(42).randn(20, 2) * 0.5

        model = FN_DBSCAN(eps=2.0, min_fuzzy_neighbors=3)
        model.fit(X)

        # Should form one cluster
        unique_labels = np.unique(model.labels_)
        # Remove noise label if present
        clusters = unique_labels[unique_labels >= 0]

        assert len(clusters) == 1
        assert model.n_clusters_ == 1


class TestFuzzyFunctionEdgeCases:
    """Test edge cases specific to fuzzy functions."""

    def test_linear_at_boundary(self):
        """Test linear function behavior at epsilon boundary."""
        X = np.array([[0, 0], [1, 0]])  # Distance = 1.0

        model = FN_DBSCAN(eps=1.0, min_fuzzy_neighbors=1.5, fuzzy_function='linear')
        model.fit(X)

        # Linear membership at d=eps is exactly 0
        # Each point: 1.0 (self) + 0.0 (other) = 1.0 < 1.5
        assert np.all(model.labels_ == -1)

    def test_exponential_at_boundary(self):
        """Test exponential function behavior at epsilon boundary."""
        X = np.array([[0, 0], [1, 0]])  # Distance = 1.0

        model = FN_DBSCAN(eps=1.0, min_fuzzy_neighbors=1.2, fuzzy_function='exponential')
        model.fit(X)

        # Exponential membership at d=eps is set to 0 (by design)
        # Each point: 1.0 (self) + 0.0 (other) = 1.0 < 1.2
        assert np.all(model.labels_ == -1)

    def test_trapezoidal_plateau_region(self):
        """Test trapezoidal function in plateau region."""
        X = np.array([[0, 0], [0.25, 0]])  # Distance = 0.25

        model = FN_DBSCAN(eps=1.0, min_fuzzy_neighbors=1.9, fuzzy_function='trapezoidal')
        model.fit(X)

        # Distance 0.25 < eps/2 = 0.5, so membership = 1.0
        # Each point: 1.0 (self) + 1.0 (other) = 2.0 >= 1.9
        assert model.n_clusters_ == 1


class TestLargeDatasets:
    """Test with larger datasets."""

    def test_moderate_size_dataset(self):
        """Test with moderate size dataset (1000 points)."""
        np.random.seed(42)
        X = np.random.randn(1000, 2)

        model = FN_DBSCAN(eps=0.5, min_fuzzy_neighbors=5)
        model.fit(X)

        assert len(model.labels_) == 1000
        assert model.n_clusters_ >= 0

    def test_high_dimensional_data(self):
        """Test with high-dimensional data."""
        np.random.seed(42)
        X = np.random.randn(100, 50)  # 50 dimensions

        model = FN_DBSCAN(eps=5.0, min_fuzzy_neighbors=3)
        model.fit(X)

        assert len(model.labels_) == 100


class TestSpecialConfigurations:
    """Test special parameter configurations."""

    def test_very_large_eps(self):
        """Test with very large epsilon."""
        X = np.array([[1, 2], [3, 4], [5, 6], [7, 8]])

        model = FN_DBSCAN(eps=100.0, min_fuzzy_neighbors=2)
        model.fit(X)

        # All points are neighbors of all, should form one cluster
        assert model.n_clusters_ == 1

    def test_very_small_min_cardinality(self):
        """Test with very small min_fuzzy_neighbors."""
        X = np.array([[1, 2], [3, 4], [5, 6]])

        model = FN_DBSCAN(eps=1.0, min_fuzzy_neighbors=0.5)
        model.fit(X)

        # Almost everything is core
        assert len(model.core_sample_indices_) >= 1

    def test_different_fuzzy_functions_same_params(self):
        """Test that different fuzzy functions can give different results."""
        X = np.array([[0, 0], [1, 0], [2, 0], [3, 0]])

        labels_linear = FN_DBSCAN(
            eps=2.0, min_fuzzy_neighbors=2.5, fuzzy_function='linear'
        ).fit_predict(X)

        labels_exp = FN_DBSCAN(
            eps=2.0, min_fuzzy_neighbors=2.5, fuzzy_function='exponential'
        ).fit_predict(X)

        labels_trap = FN_DBSCAN(
            eps=2.0, min_fuzzy_neighbors=2.5, fuzzy_function='trapezoidal'
        ).fit_predict(X)

        # All should work, though results may differ
        assert len(labels_linear) == 4
        assert len(labels_exp) == 4
        assert len(labels_trap) == 4



"""Fuzzy membership functions for FN-DBSCAN algorithm.

This module implements various fuzzy membership functions used to calculate
the fuzzy cardinality of neighborhoods in the FN-DBSCAN clustering algorithm.

References
----------
Nasibov, E. N., & Ulutagay, G. (2009). Robustness of density-based clustering
methods with various neighborhood relations. Fuzzy Sets and Systems, 160(24),
3601-3615.
"""

import numpy as np
from typing import Union


def linear_membership(distance: Union[float, np.ndarray], epsilon: float, k: float = 1.0, d_max: float = 1.0) -> Union[float, np.ndarray]:
    """Calculate linear fuzzy membership value.

    The linear membership function provides a simple linear decay from 1.0
    at distance 0 to 0.0 based on the k parameter.

    Formula: μ(d) = max(0, 1 - k·d/d_max)
    where k = d_max / ε (automatically calculated from epsilon)

    Parameters
    ----------
    distance : float or ndarray
        Distance value(s) to calculate membership for.
    epsilon : float
        Maximum neighborhood radius (must be > 0).
    k : float, default=1.0
        Parameter controlling the steepness of the membership function.
    d_max : float, default=1.0
        Maximum distance in the dataset.

    Returns
    -------
    membership : float or ndarray
        Fuzzy membership value(s) in range [0, 1].

    Examples
    --------
    >>> linear_membership(0.0, 1.0, k=1.0, d_max=1.0)
    1.0
    >>> linear_membership(0.5, 1.0, k=1.0, d_max=1.0)
    0.5
    >>> linear_membership(1.0, 1.0, k=1.0, d_max=1.0)
    0.0
    """
    if epsilon <= 0:
        raise ValueError(f"epsilon must be > 0, got {epsilon}")
    if d_max <= 0:
        raise ValueError(f"d_max must be > 0, got {d_max}")

    # Formula: μ(d) = max(0, 1 - k·d/d_max)
    membership = 1.0 - k * (distance / d_max)
    return np.clip(membership, 0.0, 1.0)


def exponential_membership(distance: Union[float, np.ndarray], epsilon: float, k: float = 20.0, d_max: float = 1.0) -> Union[float, np.ndarray]:
    """Calculate exponential fuzzy membership value.

    The exponential membership function provides a smooth exponential decay
    from 1.0 at distance 0 towards 0.0 based on the k parameter.

    Formula: μ(d) = exp(-(k·d/d_max)²)

    Parameters
    ----------
    distance : float or ndarray
        Distance value(s) to calculate membership for.
    epsilon : float
        Maximum neighborhood radius (must be > 0).
    k : float, default=20.0
        Parameter controlling the steepness of the membership function.
        Recommending k=20 for best results.
        However, dynamic calculation (k = d_max / eps) is used by default
        for better adaptability across different datasets and eps values.
    d_max : float, default=1.0
        Maximum distance in the dataset.

    Returns
    -------
    membership : float or ndarray
        Fuzzy membership value(s) in range [0, 1].

    Examples
    --------
    >>> exponential_membership(0.0, 1.0, k=1.0, d_max=1.0)
    1.0
    >>> exponential_membership(1.0, 1.0, k=1.0, d_max=1.0) < 0.5
    True
    """
    if epsilon <= 0:
        raise ValueError(f"epsilon must be > 0, got {epsilon}")
    if d_max <= 0:
        raise ValueError(f"d_max must be > 0, got {d_max}")

    # Formula: μ(d) = exp(-(k·d/d_max)²)
    membership = np.exp(-((k * distance / d_max) ** 2))

    return membership


def trapezoidal_membership(distance: Union[float, np.ndarray], epsilon: float, k: float = 1.0, d_max: float = 1.0) -> Union[float, np.ndarray]:
    """Calculate trapezoidal fuzzy membership value.

    The trapezoidal membership function has a plateau region with full
    membership (1.0) for distances up to ε/2, then linearly decays to 0.0
    at distance epsilon.

    Formula:
        μ(d) = 1.0,           if d ≤ ε/2
        μ(d) = 2(1 - d/ε),    if ε/2 < d ≤ ε
        μ(d) = 0.0,           if d > ε

    Parameters
    ----------
    distance : float or ndarray
        Distance value(s) to calculate membership for.
    epsilon : float
        Maximum neighborhood radius (must be > 0).
    k : float, default=1.0
        Parameter controlling the steepness (not used in trapezoidal).
    d_max : float, default=1.0
        Maximum distance in the dataset (not used in trapezoidal).

    Returns
    -------
    membership : float or ndarray
        Fuzzy membership value(s) in range [0, 1].

    Examples
    --------
    >>> trapezoidal_membership(0.0, 1.0)
    1.0
    >>> trapezoidal_membership(0.25, 1.0)
    1.0
    >>> trapezoidal_membership(0.75, 1.0)
    0.5
    >>> trapezoidal_membership(1.0, 1.0)
    0.0
    """
    if epsilon <= 0:
        raise ValueError(f"epsilon must be > 0, got {epsilon}")

    half_epsilon = epsilon / 2.0

    if isinstance(distance, np.ndarray):
        membership = np.ones_like(distance, dtype=float)

        # Linear decay region: ε/2 < d ≤ ε
        mask_decay = (distance > half_epsilon) & (distance <= epsilon)
        membership[mask_decay] = 2.0 * (1.0 - distance[mask_decay] / epsilon)

        # Zero region: d > ε
        mask_zero = distance > epsilon
        membership[mask_zero] = 0.0
    else:
        if distance <= half_epsilon:
            membership = 1.0
        elif distance <= epsilon:
            membership = 2.0 * (1.0 - distance / epsilon)
        else:
            membership = 0.0

    return np.clip(membership, 0.0, 1.0)


def get_fuzzy_membership_function(function_type: str):
    """Get the fuzzy membership function by name.

    Parameters
    ----------
    function_type : str
        Type of fuzzy function. Must be one of:
        - 'linear': Linear decay function
        - 'exponential': Exponential decay function
        - 'trapezoidal': Trapezoidal function with plateau

    Returns
    -------
    function : callable
        The fuzzy membership function.

    Raises
    ------
    ValueError
        If function_type is not recognized.

    Examples
    --------
    >>> func = get_fuzzy_membership_function('linear')
    >>> func(0.5, 1.0)
    0.5
    """
    functions = {
        'linear': linear_membership,
        'exponential': exponential_membership,
        'trapezoidal': trapezoidal_membership,
    }

    if function_type not in functions:
        raise ValueError(
            f"Unknown fuzzy function type: '{function_type}'. "
            f"Must be one of: {list(functions.keys())}"
        )

    return functions[function_type]


def calculate_fuzzy_cardinality(
    distances: np.ndarray,
    epsilon: float,
    fuzzy_function: str = 'linear'
) -> float:
    """Calculate the fuzzy cardinality for a set of distances.

    The fuzzy cardinality is the sum of fuzzy membership values for all
    neighbors within epsilon distance.

    Parameters
    ----------
    distances : ndarray
        Array of distances from a point to its neighbors.
    epsilon : float
        Maximum neighborhood radius (must be > 0).
    fuzzy_function : str, default='linear'
        Type of fuzzy membership function to use.

    Returns
    -------
    cardinality : float
        The fuzzy cardinality (sum of memberships).

    Examples
    --------
    >>> distances = np.array([0.0, 0.5, 1.0, 1.5])
    >>> calculate_fuzzy_cardinality(distances, epsilon=1.0, fuzzy_function='linear')
    1.5
    """
    membership_func = get_fuzzy_membership_function(fuzzy_function)
    memberships = membership_func(distances, epsilon)
    return float(np.sum(memberships))



# """Simple FN-DBSCAN clustering example with synthetic data.

# This example demonstrates the basic usage of FN-DBSCAN on a simple
# synthetic dataset with three well-separated clusters.
# """
from sklearn.datasets import make_moons
from fn_dbscan import FN_DBSCAN

X, _ = make_moons(n_samples=200, noise=0.05, random_state=42)

model = FN_DBSCAN(
    eps=0.1,                  
    min_fuzzy_neighbors=5.0,    
    min_membership=0.0,        
    fuzzy_function='exponential', 
    normalize=True              
)

labels = model.fit_predict(X)

print(f"Found {model.n_clusters_} clusters")



"""Setup configuration for fn-dbscan package."""
from setuptools import setup, find_packages

if __name__ == "__main__":
    setup(
        packages=find_packages(exclude=["tests", "tests.*", "examples", "examples.*"]),
        include_package_data=True,
    )



"""Tests for sklearn API compatibility."""

import numpy as np
import pytest
from sklearn.base import BaseEstimator, ClusterMixin
from sklearn.utils.estimator_checks import parametrize_with_checks
from sklearn.datasets import make_blobs
import pandas as pd

from fn_dbscan import FN_DBSCAN


class TestSklearnCompatibility:
    """Test sklearn API compatibility."""

    def test_inherits_base_estimator(self):
        """Test that FN_DBSCAN inherits from BaseEstimator."""
        model = FN_DBSCAN()
        assert isinstance(model, BaseEstimator)

    def test_inherits_cluster_mixin(self):
        """Test that FN_DBSCAN inherits from ClusterMixin."""
        model = FN_DBSCAN()
        assert isinstance(model, ClusterMixin)

    def test_get_params(self):
        """Test get_params method from BaseEstimator."""
        model = FN_DBSCAN(eps=0.7, min_fuzzy_neighbors=4, fuzzy_function='exponential')
        params = model.get_params()

        assert params['eps'] == 0.7
        assert params['min_fuzzy_neighbors'] == 4
        assert params['fuzzy_function'] == 'exponential'
        assert 'metric' in params

    def test_set_params(self):
        """Test set_params method from BaseEstimator."""
        model = FN_DBSCAN()

        model.set_params(eps=1.5, min_fuzzy_neighbors=10)

        assert model.eps == 1.5
        assert model.min_fuzzy_neighbors == 10

    def test_set_params_returns_self(self):
        """Test that set_params returns self."""
        model = FN_DBSCAN()
        result = model.set_params(eps=0.5)
        assert result is model

    def test_clone_compatible(self):
        """Test compatibility with sklearn.clone."""
        from sklearn.base import clone

        model1 = FN_DBSCAN(eps=0.7, min_fuzzy_neighbors=4)
        model2 = clone(model1)

        assert model2.eps == model1.eps
        assert model2.min_fuzzy_neighbors == model1.min_fuzzy_neighbors
        assert model2 is not model1

    def test_fit_method_signature(self):
        """Test that fit has correct signature."""
        import inspect

        sig = inspect.signature(FN_DBSCAN.fit)
        params = list(sig.parameters.keys())

        assert 'self' in params
        assert 'X' in params
        assert 'y' in params

    def test_fit_predict_method_signature(self):
        """Test that fit_predict has correct signature."""
        import inspect

        sig = inspect.signature(FN_DBSCAN.fit_predict)
        params = list(sig.parameters.keys())

        assert 'self' in params
        assert 'X' in params
        assert 'y' in params

    def test_labels_attribute(self):
        """Test labels_ attribute exists after fit."""
        X = np.array([[1, 2], [2, 2], [2, 3]])
        model = FN_DBSCAN()
        model.fit(X)

        assert hasattr(model, 'labels_')
        assert isinstance(model.labels_, np.ndarray)
        assert model.labels_.shape == (3,)

    def test_no_labels_before_fit(self):
        """Test that labels_ doesn't exist before fit."""
        model = FN_DBSCAN()
        assert not hasattr(model, 'labels_')


class TestDataFormats:
    """Test different input data formats."""

    def test_numpy_array_input(self):
        """Test with numpy array input."""
        X = np.array([[1, 2], [2, 2], [2, 3]])
        model = FN_DBSCAN()
        model.fit(X)

        assert model.labels_.shape == (3,)

    def test_list_input(self):
        """Test with Python list input."""
        X = [[1, 2], [2, 2], [2, 3]]
        model = FN_DBSCAN()
        model.fit(X)

        assert model.labels_.shape == (3,)

    def test_pandas_dataframe_input(self):
        """Test with pandas DataFrame input."""
        X = pd.DataFrame({
            'feature1': [1, 2, 2],
            'feature2': [2, 2, 3]
        })
        model = FN_DBSCAN()
        model.fit(X)

        assert model.labels_.shape == (3,)

    def test_pandas_series_fails(self):
        """Test that pandas Series input is handled."""
        X = pd.Series([1, 2, 3, 4, 5])
        model = FN_DBSCAN()

        # Should either work (reshape to 2D) or fail gracefully
        try:
            model.fit(X)
            assert model.labels_.shape == (5,)
        except ValueError:
            pass  # Also acceptable

    def test_float32_input(self):
        """Test with float32 input."""
        X = np.array([[1, 2], [2, 2], [2, 3]], dtype=np.float32)
        model = FN_DBSCAN()
        model.fit(X)

        assert model.labels_.shape == (3,)

    def test_int_input(self):
        """Test with integer input."""
        X = np.array([[1, 2], [2, 2], [2, 3]], dtype=int)
        model = FN_DBSCAN()
        model.fit(X)

        assert model.labels_.shape == (3,)


class TestMetrics:
    """Test different distance metrics."""

    def test_euclidean_metric(self):
        """Test with Euclidean metric."""
        X, _ = make_blobs(n_samples=50, random_state=42)
        model = FN_DBSCAN(metric='euclidean')
        model.fit(X)

        assert model.n_clusters_ >= 0

    def test_manhattan_metric(self):
        """Test with Manhattan metric."""
        X, _ = make_blobs(n_samples=50, random_state=42)
        model = FN_DBSCAN(metric='manhattan')
        model.fit(X)

        assert model.n_clusters_ >= 0

    def test_chebyshev_metric(self):
        """Test with Chebyshev metric."""
        X, _ = make_blobs(n_samples=50, random_state=42)
        model = FN_DBSCAN(metric='chebyshev')
        model.fit(X)

        assert model.n_clusters_ >= 0


class TestReproducibility:
    """Test reproducibility of results."""

    def test_multiple_fits_same_result(self):
        """Test that multiple fits give same result."""
        X, _ = make_blobs(n_samples=50, random_state=42)

        model = FN_DBSCAN(eps=2.0, min_fuzzy_neighbors=3)

        labels1 = model.fit_predict(X)
        labels2 = model.fit_predict(X)

        np.testing.assert_array_equal(labels1, labels2)

    def test_same_params_same_result(self):
        """Test that same parameters give same result."""
        X, _ = make_blobs(n_samples=50, random_state=42)

        model1 = FN_DBSCAN(eps=2.0, min_fuzzy_neighbors=3, fuzzy_function='linear')
        labels1 = model1.fit_predict(X)

        model2 = FN_DBSCAN(eps=2.0, min_fuzzy_neighbors=3, fuzzy_function='linear')
        labels2 = model2.fit_predict(X)

        np.testing.assert_array_equal(labels1, labels2)
